/*
DL2 Monitor: ./mon/dlsocket.c
Copyright (C) 2002, Rolf Jger, Ulm (Germany)

See the file README for information on usage and redistribution
of this source code file, and for a DISCLAIMER OF ALL WARRANTIES.
*/

/*----------------------------------------------------------------------------
 * Datalink Monitor (dlsocket): Datalink Receiver
 *
 * The Receiver Module implements functions which allow access to one Linux
 * packet socket. The socket is identified by the socket type and the protocol
 * type. These types are defined in the library header file (API).
 *
 * -dl2_open(st,pt)
 *  The function open a packet socket for subsequently use and returns the
 *  socket identifier, -1 on error. The socket and protocol types are given
 *  by st and pt.
 *
 * -dl2_receive(si,fd)
 *  The function receives none-truncated data from the socket opened by the
 *  dl2_open() function and writes them to the file descripted by fd. The
 *  socket is identified by si. The function only returns on error with -1.
 *  It is considered an error, if <data> can't be written completely to the
 *  output file. 
 *
 * -dl2_close(si)
 *  tbd
 *
 * Socket Type and Protocol Id Mapping:
 *   DL2_ST_* maps to SOCK_*,
 *   DL2_PT_* maps to ETH_P_*, DL2_PD_* maps to ETH_P_*.
 * 
 * Packet Output Format: <tstamp><length><packet>
 * <tstamp> 32-bit-value, ioctl-timestamp of <packet> in network byte order. 
 * <length> 16-bit-value, length of <packet> in network byte order.
 * <packet> the data received from the packet socket.
 *
 * Author: Rolf Jger,Ulm
 *----------------------------------------------------------------------------
 * YYYYMMDD  History
 * 20020408  dl2_open: network interface introduced.
 * 20020330  created from scratch
 * $Id: dlsocket.c,v 1.5 2002/04/09 06:57:52 dv Exp $
 *----------------------------------------------------------------------------
*/
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <features.h>    /* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>     /* the L2 protocols */
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>   /* The L2 protocols */
#endif
# include <unistd.h>
# include "lib/dl2mon.h"


# define STS_OPENED 0x0001
# define STS_CLOSED 0x0000
# define BUFFSIZE 0x1000
# define HIBYTE(w) (unsigned char)(((w)>>8)&0xff)
# define LOBYTE(w) (unsigned char)((w)&0xff)


static int tprotmap[] = {
                  /* Ethernet Protocol Ids*/
  ETH_P_LOOP,     /* Ethernet Loopback packet     */
  ETH_P_ECHO,     /* Ethernet Echo packet         */
  ETH_P_IP,       /* Internet Protocol packet     */
  ETH_P_X25,      /* CCITT X.25                   */
  ETH_P_ARP,      /* Address Resolution packet    */
  ETH_P_BPQ,      /* G8BPQ AX.25 Ethernet Packet  */
  ETH_P_RARP,     /* Reverse Addr Res packet      */
  ETH_P_IPX,      /* IPX over DIX                 */
  ETH_P_IPV6,     /* IPv6 over bluebook           */
  ETH_P_PPP_DISC, /* PPPoE discovery messages     */
  ETH_P_PPP_SES,  /* PPPoE session messages       */
  ETH_P_ATMMPOA,  /* MultiProtocol Over ATM       */
  ETH_P_ATMFATE,  /* Frame-based ATM Transport    */
  0
}; 
# define TPROTMAP_SIZE 13

static int dprotmap[] = {
                   /* Dummies*/
  ETH_P_802_3,    /* Dummy type for 802.3 frames  */
  ETH_P_AX25,     /* Dummy protocol id for AX.25  */
  ETH_P_ALL,      /* Every packet (be careful!!!) */
  ETH_P_802_2,    /* 802.2 frames                 */
  ETH_P_TR_802_2, /* 802.2 frames                 */
  ETH_P_CONTROL,  /* Card specific control frames */
  ETH_P_IRDA,     /* Linux-IrDA                   */
  0
};
# define DPROTMAP_SIZE 7

static int status = STS_CLOSED;
static int check_socket_type(int);
static int check_protocol(int);
static int bind_interface(int,int,const char*);

//-exports--------------------------------------------------------------------
int dl2_open(int stype,int protocol,const char* ifname) {
  if (status == STS_CLOSED) {
    int sockfd = DL2_FAILURE;
    if ((stype = check_socket_type(stype)) < 0) return DL2_FAILURE;
    if ((protocol = check_protocol(protocol)) < 0) return DL2_FAILURE;
    sockfd = socket(PF_PACKET,stype,htons(protocol));
    if (sockfd >= 0) {
      if (ifname != NULL) {
        int rc = bind_interface(sockfd,htons(protocol),ifname);
        if (rc < 0) {close(sockfd); return DL2_FAILURE;}
      }
      status = STS_OPENED;
    }
    return sockfd;
  } return DL2_FAILURE;
} //dl2_open


int dl2_close(int sockfd) {
  close(sockfd);
  status = STS_CLOSED;
  return 0;
} //dl2_close


int dl2_receive(int sockfd,int outfd) {
  unsigned char *buffer = (unsigned char*)malloc(BUFFSIZE);
  if (buffer != NULL) {
    int rc = 0, insize, outsize, n;
    unsigned char *pbuffer = buffer+DL2_TIME_SIZE+DL2_LENGTH_SIZE;
    const int plength = BUFFSIZE-DL2_TIME_SIZE-DL2_LENGTH_SIZE;
    for (;rc >= 0 && status&STS_OPENED;) {
      insize = (rc = recv(sockfd,pbuffer,plength,MSG_TRUNC));
      if (insize > 0) {
        struct timeval timestamp;
        long tv_sec;
        int psize = insize;
	ioctl(sockfd,SIOCGSTAMP,&timestamp);
        tv_sec = timestamp.tv_sec;
        for (n = DL2_TIME_SIZE-1; n >= 0; n--) {
          buffer[n] = (unsigned char)(tv_sec & 0xff); tv_sec >>= 8;
        } insize += DL2_TIME_SIZE;
        n = DL2_TIME_SIZE;
        for (n += DL2_LENGTH_SIZE-1; n >= DL2_TIME_SIZE; n--) {
          buffer[n] = (unsigned char)(psize & 0xff); psize >>= 8;
        } insize += DL2_LENGTH_SIZE;
        outsize = (rc = write(outfd,buffer,insize));
        if (outsize < insize) rc = -1; 
      }
    } free(buffer); 
    return rc;
  } return DL2_FAILURE;
} //dl2_receive


//-locals---------------------------------------------------------------------
/* return -1 if not supported, 0 otherwise; 
 */
int check_socket_type(int st) {
  if (st == DL2_ST_RAW) return SOCK_RAW;
  if (st == DL2_ST_DGRAM) return SOCK_DGRAM;
  return -1;
} //check_socket_type


/* return -1 if not supported, 0 otherwise;
 * supported protocols form a subset of those allowed by the ethn devices, 
 * see <linux/if_ether.h> for details.
 */
int check_protocol(int p) {
  if (0 <= p && p < TPROTMAP_SIZE) return tprotmap[p];
  p -= 256;
  if (0 <= p && p < DPROTMAP_SIZE) return dprotmap[p];
  return -1; 
} //check_protocol


int bind_interface(int fd,int protocol,const char *ifname) {
  struct ifreq req;
  struct sockaddr_ll sa;
  int rc;
  strncpy(req.ifr_name,ifname,sizeof(req.ifr_name)-1);
  if ((rc = ioctl(fd,SIOCGIFINDEX,&req)) >= 0) { 
    sa.sll_family = AF_PACKET;
    sa.sll_protocol = protocol;
    sa.sll_ifindex = req.ifr_ifindex;
    rc = bind(fd,(struct sockaddr*)&sa,sizeof(sa));
  } return rc;
} //bind_interface

/***EOM***/
