/*
DL2 Monitor: ./pa/palcp.c
Copyright (C) 2002, Rolf Jger, Ulm (Germany)

See the file README for information on usage and redistribution
of this source code file, and for a DISCLAIMER OF ALL WARRANTIES.
*/

/*----------------------------------------------------------------------------
 * Datalink Monitor (lcp): Protocol Analysis (RFC1661)
 * PPP Link Control Protocol
 *
 * P_*  indicate frame positions of certain fields;
 * C_*  indicate known LCP codes;
 * CO_* indicate known configuration options
 * default contains: code, request-id
 *
 * Configuration options have been assigned by 
 *   RFC1661 (specification of LCP), 
 *   RFC1700 (latest assigned numbers)
 *   ppp-numbers (iana online document)
 * They are part of packets of code Configure-Request/Ack/Nak/Reject, only.
 *
 * The content of some LCP packets will not be further refined since it has
 * special meaning. These packets are Terminate-Request/Ack, Code-Reject, and
 * Protocol-Reject.
 * 
 * Author: Rolf Jger,Ulm
 *----------------------------------------------------------------------------
 * YYYYMMDD  History
 * 20020427  length instead of packet-length
 * 20020405  fully implemented.
 * 20020404  implemented from RFC1661 (without option analysis) 
 * 20020330  created from scratch. 
 * $Id: palcp.c,v 1.6 2002/04/29 12:59:27 dv Exp $
 *----------------------------------------------------------------------------
*/
# include <stdio.h>
# include <string.h>
# include "lib/dl2pa.h"
# include "pa/lcp.h"

# define NSTR  ""

// LCP frame positions:
# define P_CODE  0x00
# define P_REQUESTID  0x01
# define P_LENGTH  0x02
# define P_DATA  0x04

// known codes: 1..d (RFC1700) 
# define C_Crequest  0x01
# define C_Cack      0x02
# define C_Cnak      0x03
# define C_Creject   0x04
# define C_Trequest  0x05
# define C_Tack      0x06
# define C_Cdreject  0x07
# define C_Preject   0x08
# define C_Erequest  0x09
# define C_Ereply    0x0a
# define C_Drequest  0x0b
# define C_Identify  0x0c
# define C_Timerest  0x0d

// default indications:
static const char* codestr[] = {
  "",
  "ConfRequest",
  "ConfAck",
  "ConfNak",
  "ConfReject",
  "TermRequest",
  "TermAck",
  "CodeReject",
  "ProtReject",
  "EchoRequest",
  "EchoReply",
  "DiscRequest",
  "Identification",
  "TimeRemaining"
};
# define MAXCODE C_Timerest

// known configuration options:  RFC1661
# define CO_MRU            0x01
# define CO_Asyncmap	   0x02
# define CO_AuthProtocol   0x03
# define CO_QualProtocol   0x04
# define CO_MagicNumber    0x05
# define CO_PfieldComp     0x07
# define CO_ACfiledComp    0x08

struct OptionFM {const char* F; int M;};
# define MAXOPT 8
struct OptionFM config[MAXOPT+1];
# define ADDOPT(i,t) {config[i].F=LCP_F##t;config[i].M=LCP_M##t;}

static int init_tables = 1;
static const int is_config_code = (1<<C_Crequest)|(1<<C_Cack)|(1<<C_Cnak)|(1<<C_Creject);
static const int is_echo_code = (1<<C_Erequest)|(1<<C_Ereply);
static const int is_other_code = (1<<C_Drequest);
static int print_option(Octet*,int);

//-exports--------------------------------------------------------------------
int dl2_palcp(Dl2_Frame *frame,int verbose) {
  Octet *buffer = frame->buffer;
  int length = frame->length, i = 0;
  Octet code;
  Doctet pl;

  if (init_tables) { 
    int j;
    for (j = 0; j <= MAXOPT; j++) {config[j].F = NSTR; config[j].M = 0;}
    ADDOPT(CO_MRU,mru);
    ADDOPT(CO_Asyncmap,accm);
    ADDOPT(CO_AuthProtocol,auth_protocol);
    ADDOPT(CO_QualProtocol,qual_protocol);
    ADDOPT(CO_MagicNumber,magic_number);
    ADDOPT(CO_PfieldComp,pf_compression);
    ADDOPT(CO_ACfiledComp,acf_compression);
    init_tables = 0;
  }

  // check for required packet length:
  if (length < P_DATA) return -1;

  //code:
  code = buffer[i];
  if (verbose & LCP_Mcode) {
    printf("%s=%02x",LCP_Fcode,buffer[i]);
  } 
  else if (verbose & LCP_Mdefault) {
    const char *cs = (code <= MAXCODE ? codestr[code] : PA_FC_UNKNOWN); 
    printf(" %s",cs); 
  } i = P_REQUESTID;

  // request-id:
  if (verbose & LCP_Mrequest_id) {
    printf(" %s=%02x",LCP_Frequest_id,buffer[i]);
  }
  else if (verbose & LCP_Mdefault) {
    printf("(%02x)",buffer[i]);
  } i = P_LENGTH;

  // packet-length:
  if (verbose & LCP_Mlength) {
    printf(" %s=",LCP_Flength);
    for (; i < P_DATA; i++) printf("%02x",buffer[i]);
  } else i = P_DATA;
  
  // data:
  if ((pl = str2doct(buffer+P_LENGTH)) > length) pl = length;
  if (verbose & LCP_Mdata) {
    printf(" %s=",LCP_Fdata);
    for (; i < pl; i++) printf("%02x",buffer[i]);
  }

  // options:
  if ((1<<code) & is_config_code) {
    if (verbose & LCP_Moptions) {
      for (i = P_DATA; i < pl;) {
        Octet type = buffer[i];
        if (type <= MAXOPT) {
          printf(" %s=",config[type].F);
          i += print_option(buffer+i,pl);
        }
        else {
          printf(" %s=",PA_FT_UNKNOWN);
          break;
        }
      }
    }
    else for (i = P_DATA; i < pl;) {
      Octet type = buffer[i];
      if (type <= MAXOPT) {
        if (config[type].M & verbose) {
          printf(" %s=",config[type].F);
          i += print_option(buffer+i,pl);
        }
        else {i += buffer[i+1]; continue;}
      }
      else {
        printf(" %s=",PA_FT_UNKNOWN);
        break;
      }
    }
  }

  // echo request/reply:
  else if ((1<<code) & is_echo_code) {
    if (verbose & LCP_Mecho_data) {
      printf(" %s=",LCP_Fecho_data);
      for (i = P_DATA; i < pl; i++) printf("%02x",buffer[i]);
    }
  }

  // others:
  else if ((1<<code) & is_other_code) {
    if (verbose & LCP_Mother_data) {
      printf(" %s=",LCP_Fother_data);
      for (i = P_DATA; i < pl; i++) printf("%02x",buffer[i]);
    }
  }
  return i;  
} //dl2_palcp


//-locals---------------------------------------------------------------------
/* print one option, return number of analyzed octets, ml on error;
 * option format: <type 1><length 1><data ?>;
 * <length> indicates length including <type> and <data>.
*/
inline int print_option(Octet *buffer,int ml) {
  Octet type = buffer[0], length = buffer[1];
  int i = ml;
  if (length >= 2) {  
    if (length > ml) length = ml;
    for (i = 2; i < length; i++) printf("%02x",buffer[i]);
  } return i;
} //print_option

/***EOM***/
